# frozen_string_literal: true

require_relative 'soql_objects'

module LeapSalesforce
  # Generators for creating code
  module Generator
    # Creates Soql Enumerations / Picklists from objects in LeapSalesforce.objects_to_verify
    class SoqlEnums
      include Generator

      # @return [String] Folder where enumeration objects are stored
      ENUM_FOLDER = File.join(LeapSalesforce.lib_folder, 'metadata', 'enum').freeze

      # Create Soql enumerations for all objects specified to verify
      def create
        if LeapSalesforce.objects_to_verify.empty?
          raise LeapSalesforce::SetupError, 'LeapSalesforce.objects_to_verify is empty. ' \
        'Please set the list of objects you want to verify in .leap_salesforce:soql_objects ' \
        'and then run task "leaps:create_soql_objects" to create objects'
        end
        reset_enum_folder
        LeapSalesforce.objects_to_verify.each { |entity| create_picklists_for entity }
        cleanup_files_created
      end

      # Create files for each picklist found for that entity
      # @param [LeapSalesforce::SoqlData] entity An object representing an object in Salesforce
      #   Object inheriting from SoqlData that has picklists underneath it
      def create_picklists_for(entity)
        soql_object = LeapSalesforce.soql_objects.find { |so| so.class_name == entity.to_s }
        raise "Could not find soql object for '#{entity}'" unless soql_object

        unless soql_object.create_enum != false
          puts "Skipping picklists for #{entity}"
          return
        end
        puts "Creating picklists for #{entity}"
        @entity_name = entity
        picklists = entity.picklists
        puts "#{picklists.count} picklists found"
        picklists.each do |picklist|
          if soql_object.excludes?(picklist)
            puts "Excluding picklist '#{picklist}'"
          else
            generate_picklist_file picklist
          end
        end
      end

      # Generate file for a picklist from it's metadata
      def generate_picklist_file(picklist)
        @picklist = picklist
        @enum_name = picklist.to_class_name
        picklist_name = picklist.to_key_name
        values = @entity_name.picklist_for(picklist)
        @enum_values = {}
        values.each do |value|
          @enum_values[value.to_key_name_for(@enum_values)] = value
        end
        content = read_template 'picklist.rb.erb', binding
        file = File.join(ENUM_FOLDER, @entity_name.to_s.snakecase, "#{picklist_name}.rb")
        generate_file file, content
      end

      # Reset state of Enum folder so that enums that no longer exist do not persist
      def reset_enum_folder
        FileUtils.rm_rf ENUM_FOLDER
      end

      # Clean up files generated for all picklists
      def cleanup_files_created
        `rubocop -A #{ENUM_FOLDER} --display-only-fail-level-offenses --enable-pending-cops`
      end
    end
  end
end
